#!/usr/bin/python
#
# Project: lib Instrument Patch Python testing suite
# File:    Testing.py
# Descr:   Testing support functions used by other tests
# Author:  Element Green <element@elementsofsound.org>
# Date:    2016-07-01
# License: Public Domain
#
import sys
import os
import argparse
import gi
from gi.repository import GLib
from gi.repository import GObject

gi.require_version ('Ipatch', '1.1')
from gi.repository import Ipatch

ErrorCount = 0
LogMask = 0
Verbose = 1
TEST_DATA_PATH = "test-data/"

def createArgParser (description):
  parser = argparse.ArgumentParser (description=description)
  parser.add_argument ('-v', '--verbose', action='count', help='Increase verbosity')
  parser.add_argument ('--quiet', action='store_true', help='Only display errors')
  return parser

def parseArgs (parser):
  global Verbose

  args = parser.parse_args ()

  if args.verbose > 0:
    Verbose = args.verbose + 1

  if args.quiet:
    Verbose = 0

  return args

def LogTrap (log_domain, log_level, message, user_data):
  global LogMask

  LogMask |= log_level
  GLib.log_default_handler ("Testing", log_level, message, 0)

def exit ():
  if LogMask: print ("GLib warning/error messages occurred!")
  if ErrorCount > 0: print ("%d Errors occurred!" % ErrorCount)
  sys.exit (0 if ErrorCount == 0 and LogMask == 0 else 1)

def info (s):
  if Verbose >= 2:
    print (s)

def msg (s):
  if Verbose >= 1:
    print (s)

def error (s):
  global ErrorCount

  print (s)
  ErrorCount += 1

def compareItems (item1, item2, recurse=True, blackList=()):
  """Compare two Ipatch.Item objects"""

  if type (item1) != type (item2):
    raise RuntimeError ('Item type %s != %s' % (str (type (item1)), str (type (item2))))

  typename = str (type (item1))
  props = sorted ([prop for prop in item1.list_properties ()], key=lambda prop: prop.name)

  blprops = [propname for typ, propname in blackList if isinstance (item1, typ)]

  for p in props:
    if p.name in blprops: continue

    val1 = item1.get_property (p.name)
    val2 = item2.get_property (p.name)

    valtype = p.value_type

    if val1 is None != val2 is None:
      raise RuntimeError ('Property value None mismatch for property %s:%s'
                          % (typename, p.name))

    if valtype.is_a (GObject.Object):
      if val1 is not None and recurse:
        compareItems (val1, val2, recurse=False, blackList=blackList)   # Don't recurse into object fields (causes recursion loops)
    elif valtype.is_a (Ipatch.Range):
      if val1.low != val2.low or val1.high != val2.high:
        RuntimeError ('Range value mismatch for property %s:%s, %d-%d != %d-%d'
                      % (typename, p.name, val1.low, val1.high, val2.low, val2.high))
    elif valtype.is_a (GObject.GBoxed):
      if valtype.name == 'IpatchSF2ModList':
        if len (val1) != len (val2):
          RuntimeError ('Modulator list length mismatch for property %s:%s, %d != %d'
                        % (typename, p.name, len (val1), len (val2)))

        for i in xrange (0, len (val1)):
          v1 = val1[i]
          v2 = val2[i]

          v1 = (v1.src, v1.amtsrc, v1.dest, v1.amount, v1.trans)
          v2 = (v2.src, v2.amtsrc, v2.dest, v2.amount, v2.trans)

          if v1 != v2:
            RuntimeError ('Modulator value mismatch for property %s:%s, %s != %s'
                          % (typename, p.name, str (v1), str (v2)))
      else:
        print ('Ignoring property %s:%s of type %s'
               % (typename, p.name, str (valtype)))
    elif val1 != val2:
      raise RuntimeError ('Value mismatch for property %s:%s, %s != %s'
                          % (typename, p.name, str (val1), str (val2)))

  if recurse and isinstance (item1, Ipatch.Container):
    children1 = item1.get_children ()
    children2 = item2.get_children ()

    if len (children1) != len (children2):
      raise RuntimeError ('Container of type %s has unequal children count, %d != %d'
                          % (typename, len (children1), len (children2)))

    for i in xrange (0, len (children1)):
      compareItems (children1[i], children2[i], blackList=blackList)

# Log flags we want to trap on
flags = GLib.LogLevelFlags.LEVEL_WARNING | GLib.LogLevelFlags.LEVEL_ERROR \
 | GLib.LogLevelFlags.LEVEL_CRITICAL | GLib.LogLevelFlags.FLAG_FATAL | GLib.LogLevelFlags.FLAG_RECURSION

GLib.log_set_handler (None, flags, LogTrap, None)

# Create test data directory if needed
try: 
  os.makedirs (TEST_DATA_PATH)
except OSError:
  if not os.path.isdir (TEST_DATA_PATH):
    raise




#!/usr/bin/python
#
# Project: lib Instrument Patch Python testing suite
# File:    SF2.py
# Descr:   SoundFont tests
# Author:  Element Green <element@elementsofsound.org>
# Date:    2016-07-01
# License: Public Domain
#
import Test
import gi
import ctypes
import math
import sys
import os

from gi.repository import GObject

gi.require_version ('Ipatch', '1.1')
from gi.repository import Ipatch

# Sample audio format (native SoundFont 16 bit mono signed little endian byte order audio)
SAMPLE_FORMAT = Ipatch.SampleWidth.BIT16 | Ipatch.SampleChannel.MONO | Ipatch.SampleSign.SIGNED | Ipatch.SampleEndian.LENDIAN

# Format used for writing sample data (converted from this format)
ACCESS_FORMAT = Ipatch.SampleWidth.FLOAT | Ipatch.SampleChannel.MONO | Ipatch.SampleSign.SIGNED
SAMPLE_RATE             = 48000         # Sample rate
SAMPLE_LOOP_PADDING     = 4             # Number of samples before and after the sample loop
TEST_SF2_FILENAME       = os.path.join (Test.TEST_DATA_PATH, 'test.sf2')

# Set sample access format endian byte order depending on host system
ACCESS_FORMAT |= Ipatch.SampleEndian.BENDIAN if sys.byteorder == 'big' else Ipatch.SampleEndian.LENDIAN

def SineWave (periodPos):
  """Sine waveform generation function"""
  return math.sin (periodPos * 2 * math.pi)

def SawWave (periodPos):
  """Triangle waveform generation function"""
  if (periodPos <= 0.25): return periodPos / 0.25
  elif (periodPos <= 0.75): return (0.75 - periodPos) * 4.0 - 1.0
  else: return (1.0 - periodPos) * -4.0

def SquareWave (periodPos):
  """Square waveform generation function"""
  return 1.0 if periodPos <= 0.5 else 0.0

def sampleDataGen (freq, mathfunc):
  """Raw sample data generator"""

  period = float (SAMPLE_RATE) / freq

  start = int (math.floor (period - SAMPLE_LOOP_PADDING))
  end = int (math.ceil (2 * period + SAMPLE_LOOP_PADDING))
  size = end - start

  buf = (ctypes.c_float * size)()

  # Generate a single period of the waveform, including loop padding
  for i in xrange (0, size):
    periodPos = math.fmod (start + i, period) / period
    buf[i] = mathfunc (periodPos)

  return bytearray (buf)

def createSF2Sample (floatData):
  """Generate sample data object"""

  sampleSize = len (floatData) / 4

  # Create RAM sample store then set format, rate, and size
  store = Ipatch.SampleStoreRam ()
  store.set_properties (sample_format=SAMPLE_FORMAT, sample_rate=SAMPLE_RATE, sample_size=sampleSize)

  # Open a handle to the sample store and write the waveform floating point data to it
  retval, handle = store.handle_open ('w', ACCESS_FORMAT, 0)
  handle.write (0, floatData)   # 0 is the position to write to (start of sample)
  handle.close ()

  # Create sample data object and add sample store to it
  sampleData = Ipatch.SampleData ()
  sampleData.add (store)

  # Create SF2Sample and add sample data to it
  sf2Sample = Ipatch.SF2Sample ()
  sf2Sample.set_data (sampleData)

  loopStart = SAMPLE_LOOP_PADDING
  loopEnd = sampleSize - SAMPLE_LOOP_PADDING - 1

  # Calculate root note and fine tune based on loop length for a single waveform
  notecents = Ipatch.unit_hertz_to_cents (float (SAMPLE_RATE) / (loopEnd - loopStart))
  cents = math.fmod (notecents, 100.0)

  # Fine tune offset corrects for inaccuracy of a single digital waveform
  if cents <= 50.0:
    rootNote = int (notecents / 100.0)
    fineTune = -int (round (cents))
  else:
    rootNote = int (notecents / 100.0) + 1
    fineTune = int (round (100.0 - cents))

  sf2Sample.set_properties (sample_rate=SAMPLE_RATE, loop_start=loopStart, loop_end=loopEnd,
                            root_note=rootNote, fine_tune=fineTune)
  return sf2Sample

def addSF2Preset (sf2, name, midiProgram, sampleMathFunc):
  """Create a SoundFont preset, with an instrument, and containing several generated samples"""

  # Create a SF2 preset object, set its MIDI bank and program numbers (MIDI locale), and add it to the SF2 object
  preset = Ipatch.SF2Preset ()
  preset.set_name (name)
  preset.set_midi_locale (0, midiProgram)
  sf2.add (preset)

  # Create a SF2 instrument object and add it to the SF2 object
  inst = Ipatch.SF2Inst ()
  inst.set_name (name)
  sf2.add (inst)
  preset.new_zone (inst)        # Create new preset zone and add instrument to it

  # Create 2 modulators to disable the default pitch modulation control and modulate the filter cutoff instead
  mod = Ipatch.SF2Mod ()
  mod.src = 1 | Ipatch.SF2ModControlPalette.MIDI | Ipatch.SF2ModDirection.POSITIVE \
    | Ipatch.SF2ModPolarity.UNIPOLAR | Ipatch.SF2ModType.LINEAR
  mod.dest = Ipatch.SF2GenType.VIB_LFO_TO_PITCH
  mod.amount = 0
  mod.amtsrc = 0
  mod.trans = Ipatch.SF2ModTransform.LINEAR

  modlist = [mod]       # Create modulator list

  mod = Ipatch.SF2Mod ()
  mod.src = 1 | Ipatch.SF2ModControlPalette.MIDI | Ipatch.SF2ModDirection.POSITIVE \
    | Ipatch.SF2ModPolarity.UNIPOLAR | Ipatch.SF2ModType.LINEAR
  mod.dest = Ipatch.SF2GenType.MOD_LFO_TO_FILTER_CUTOFF
  mod.amount = 2400
  mod.amtsrc = 0
  mod.trans = Ipatch.SF2ModTransform.LINEAR

  modlist.append (mod)  # Append 2nd modulator to list

  # Set instrument global modulators to modulator list
  inst.set_mods (modlist)

  # Create A note samples for octaves A0 through A6 (7 octaves) and add to instrument zones
  for octave in xrange (0, 7):
    note = 33 + octave * 12     # MIDI note 33 is A0
    sample = createSF2Sample (sampleDataGen (Ipatch.unit_cents_to_hertz (note * 100), sampleMathFunc))
    sample.set_name ('%s A%d' % (name, octave))
    sf2.append (sample)         # Add sample to SF2 base object

    # Create instrument zone
    izone = Ipatch.SF2IZone ()
    inst.add (izone)

    izone.set_sample (sample)   # Assign the sample to the zone

    # Calculate optimal note ranges for samples
    start = note - 5 if octave > 0 else 0
    end = note + 6 if octave < 6 else 127

    izone.set_properties (loop_type=Ipatch.SampleLoopType.STANDARD,             # Enable looping
                          note_range=Ipatch.Range.new (start, end))             # Create range and assign to note range

def createSF2 ():
  """Create SoundFont object"""

  # Create SoundFont object, a file object, and assign it
  sf2 = Ipatch.SF2 ()
  sf2.props.name = "Test SoundFont"
  sf2.props.author = "Robo Python 3000+"
  sf2.props.comment = """This is a test SoundFont generated by the Instrument Patch library Python binding.
One should find 3 different presets for Sine, Saw, and Square waves with 7 samples each.
Enjoy!"""

  sf2File = Ipatch.SF2File ()
  sf2File.set_name (TEST_SF2_FILENAME)
  sf2.set_file (sf2File)

  # Add presets for sine, saw, and square waves
  addSF2Preset (sf2, "Sine", 0, SineWave)
  addSF2Preset (sf2, "Saw", 1, SawWave)
  addSF2Preset (sf2, "Square", 2, SquareWave)

  return sf2


# Main
if __name__ == "__main__":
  parser = Test.createArgParser ('libInstPatch SoundFont tests')
  parser.add_argument ('-n', '--nodelete', action='store_true', help="Don't delete created test SoundFont (%s)" % TEST_SF2_FILENAME)
  args = Test.parseArgs (parser)

  Ipatch.init ()

  Test.msg ("Creating SoundFont")
  sf2 = createSF2 ()

  Test.msg ("Saving SoundFont")
  sf2.save ()

  Test.msg ("Loading SoundFont")
  sf2file = Ipatch.SF2File ()
  sf2file.set_property ('file-name', TEST_SF2_FILENAME)
  cmpsf2 = Ipatch.convert_object_to_type (sf2file, Ipatch.SF2)

  blacklistProps = {
    (Ipatch.SF2, "software"),             # SF2 software will include the loaded software version appended
    (Ipatch.SF2File, "sample-size")       # Sample size is the total sample data in the SoundFont file which will differ
  }

  # Clear changed and saved flags so they don't trigger a value mismatch
  sf2.clear_flags (Ipatch.BaseFlags.CHANGED | Ipatch.BaseFlags.SAVED)
  cmpsf2.clear_flags (Ipatch.BaseFlags.CHANGED | Ipatch.BaseFlags.SAVED)

  Test.msg ("Verifying SoundFont")
  Test.compareItems (sf2, cmpsf2, blackList=blacklistProps)

  if not args.nodelete:
    os.unlink (TEST_SF2_FILENAME)

  Test.exit ()




#!/usr/bin/python

# Python example of creating a SoundFont from multiple audio samples
#
# Updated on July 22, 2006
# Now uses more convenient converter functions.
#
# Josh Green - Sep 13, 2004
# Use this example as you please (public domain)
#
# Note that the Python binding API is not stable yet and may change
#

import ipatch
import gobject, sys

if len (sys.argv[1:]) == 0:
  print "Usage: create_sf2.py sample1.wav [sample2.wav sample3.aiff ...]"
  sys.exit ()

sf2 = ipatch.SF2 ()		# Create new SoundFont object

# For each file specified on command line..
for fname in sys.argv[1:]:
  # Identify and open the file
  f = ipatch.ipatch_file_identify_open (fname, "r")

  if not f:
    print "Failed to identify file '%s'" % fname
    continue

  # Convert sample file to SF2Sample object
  sample = ipatch.ipatch_convert_object_to_type (f, ipatch.SF2Sample)
  if not sample: continue

  sf2.add_unique (sample)		# Add sample and ensure name is unique

  name = sample.get_property ("name")	# Get the sample name

  inst = ipatch.SF2Inst ()		# Create new SoundFont instrument
  inst.set_property ("name", name)	# Set the instrument name = sample name
  inst.new_zone (sample)		# Add new zone to inst => sample
  sf2.add_unique (inst)			# Add inst and ensure name is unique

  preset = ipatch.SF2Preset ()		# Create new preset
  preset.set_property ("name", name)	# Set preset name = sample name
  preset.new_zone (inst)		# Add zone to preset => inst
  sf2.add_unique (preset)		# Add preset, name/bank/preset # unique

# Create SoundFont file object, set its name and open for writing
sffile = ipatch.SF2File ()
sffile.open ("output.sf2", "w")

ipatch.ipatch_convert_objects (sf2, sffile)

sffile.close ()



#!/usr/bin/python
#
# Project: lib Instrument Patch Python testing suite
# File:    Converter.py
# Descr:   Object converter tests
# Author:  Element Green <element@elementsofsound.org>
# Date:    2016-07-04
# License: Public Domain
#
import Test
import gi

from gi.repository import GObject

gi.require_version ('Ipatch', '1.1')
from gi.repository import Ipatch

# Common converter source to destination type mappings
CONVERTER_TYPES = (
  ("IpatchFile", "IpatchBase"),
  ("IpatchBase", "IpatchFile"),
  ("IpatchItem", "IpatchSF2VoiceCache"),
  ("IpatchSndFile", "IpatchItem")
)

# Main
if __name__ == "__main__":
  parser = Test.createArgParser ('libInstPatch Converter tests')
  args = Test.parseArgs (parser)

  Ipatch.init ()

  convinfo = [Ipatch.get_converter_info (t) for t in Ipatch.find_converters (GObject.TYPE_NONE, GObject.TYPE_NONE, 0)]
  convinfo.sort (key=lambda info: GObject.type_name (info.conv_type))

  if len (convinfo) == 0:
    Test.error ("No converters found!")

  Test.msg ("Found %d converters\n" % len (convinfo))

  summaryCounts = [0] * len (CONVERTER_TYPES)
  summaryOther = 0

  # Show converter details
  for convindex in xrange (0, len (convinfo)):
    conv = convinfo[convindex]

    s = GObject.type_name (conv.conv_type) + "\n"

    if conv.src_count == Ipatch.ConverterCount.ONE_OR_MORE: count = '+'
    elif conv.src_count == Ipatch.ConverterCount.ZERO_OR_MORE: count = '*'
    else: count = str (conv.src_count)

    if conv.src_match == GObject.TYPE_INVALID:
      s += "  src[%s]=%s " % (count, GObject.type_name (conv.src_type))
    else: s += "  src[%s]=%s - %s" \
      % (count, GObject.type_name (conv.src_match), GObject.type_name (conv.src_type))

    if conv.flags & Ipatch.ConverterFlags.SRC_DERIVED:
      s += "+ DESCENDANTS\n"
    else: s += "\n"

    if conv.dest_count == Ipatch.ConverterCount.ONE_OR_MORE: count = '+'
    elif conv.dest_count == Ipatch.ConverterCount.ZERO_OR_MORE: count = '*'
    else: count = str (conv.dest_count)

    if conv.dest_match == GObject.TYPE_INVALID:
      s += "  dst[%s]=%s" % (count, GObject.type_name (conv.dest_type))
    else: s += "  dst[%s]=%s - %s" \
      % (count, GObject.type_name (conv.dest_match), GObject.type_name (conv.dest_type))

    if conv.flags & Ipatch.ConverterFlags.DEST_DERIVED:
      s += "+ DESCENDANTS"

    Test.info (s)

    for i in xrange (0, len (CONVERTER_TYPES)):
      convTypes = CONVERTER_TYPES[i]

      if GObject.type_is_a (conv.src_type, GObject.type_name (convTypes[0])) \
          and GObject.type_is_a (conv.dest_type, GObject.type_name (convTypes[1])):
        summaryCounts[i] += 1
        break
    else: summaryOther += 1

  Test.info ("")

  Test.msg ("Converter type summary:")

  for i in xrange (0, len (CONVERTER_TYPES)):
    convTypes = CONVERTER_TYPES[i]
    Test.msg ("%s -> %s: %d" % (convTypes[0], convTypes[1], summaryCounts[i]))

  if summaryOther > 0:
    Test.msg ("Other: %d" % (convTypes[0], convTypes[1], summaryCounts[i]))

  Test.exit ()




#!/usr/bin/python
#
# Project: lib Instrument Patch Python testing suite
# File:    SF2.py
# Descr:   SoundFont tests
# Author:  Element Green <element@elementsofsound.org>
# Date:    2016-07-01
# License: Public Domain
#
"""
Convert between all possible audio format pairs using the following steps:
- Convert double to first format of format pair
- Convert from first format to second format
- Convert second format back to double
- Calculate maximum audio sample difference between new double audio and original
"""

import Test
import gi
import ctypes
import math
import sys

from gi.repository import GObject

gi.require_version ('Ipatch', '1.1')
from gi.repository import Ipatch

# Default values are non powers of 2 to decrease the chance of masking periodic data issues
DEFAULT_SAMPLE_SIZE             = 1756          # default test waveform size in samples
DEFAULT_SAMPLE_PERIOD           = 123           # default waveform period in samples
DEFAULT_TRANSFORM_SIZE          = 65000         # default transform buffer size in bytes
MAX_DIFF_ALLOWED                = 0.016         # maximum difference allowed

# all available sample format combinations
testFormats = (
  Ipatch.SampleWidth.BIT8,
  Ipatch.SampleWidth.BIT16,
  Ipatch.SampleWidth.BIT24,
  Ipatch.SampleWidth.BIT32,
  Ipatch.SampleWidth.FLOAT,
  Ipatch.SampleWidth.DOUBLE,
  Ipatch.SampleWidth.REAL24BIT,

  Ipatch.SampleWidth.BIT8 | Ipatch.SampleSign.UNSIGNED,
  Ipatch.SampleWidth.BIT16 | Ipatch.SampleSign.UNSIGNED,
  Ipatch.SampleWidth.BIT24 | Ipatch.SampleSign.UNSIGNED,
  Ipatch.SampleWidth.BIT32 | Ipatch.SampleSign.UNSIGNED,
  Ipatch.SampleWidth.REAL24BIT | Ipatch.SampleSign.UNSIGNED,

  Ipatch.SampleWidth.BIT16 | Ipatch.SampleEndian.BENDIAN,
  Ipatch.SampleWidth.BIT24 | Ipatch.SampleEndian.BENDIAN,
  Ipatch.SampleWidth.BIT32 | Ipatch.SampleEndian.BENDIAN,
  Ipatch.SampleWidth.FLOAT | Ipatch.SampleEndian.BENDIAN,
  Ipatch.SampleWidth.DOUBLE | Ipatch.SampleEndian.BENDIAN,
  Ipatch.SampleWidth.REAL24BIT | Ipatch.SampleEndian.BENDIAN,

  Ipatch.SampleWidth.BIT16 | Ipatch.SampleSign.UNSIGNED | Ipatch.SampleEndian.BENDIAN,
  Ipatch.SampleWidth.BIT24 | Ipatch.SampleSign.UNSIGNED | Ipatch.SampleEndian.BENDIAN,
  Ipatch.SampleWidth.BIT32 | Ipatch.SampleSign.UNSIGNED | Ipatch.SampleEndian.BENDIAN,
  Ipatch.SampleWidth.REAL24BIT | Ipatch.SampleSign.UNSIGNED | Ipatch.SampleEndian.BENDIAN
)

testFormatNames = (
  "8bit-signed",
  "16bit-signed-lendian",
  "24bit-signed-lendian",
  "32bit-signed-lendian",
  "float-lendian",
  "double-lendian",
  "real24bit-signed-lendian",

  "8bit-unsigned",
  "16bit-unsigned-lendian",
  "24bit-unsigned-lendian",
  "32bit-unsigned-lendian",
  "real24bit-unsigned-lendian",

  "16bit-signed-bendian",
  "24bit-signed-bendian",
  "32bit-signed-bendian",
  "float-bendian",
  "double-bendian",
  "real24bit-signed-bendian",

  "16bit-unsigned-bendian",
  "24bit-unsigned-bendian",
  "32bit-unsigned-bendian",
  "real24bit-unsigned-bendian"
)


parser = Test.createArgParser ('libInstPatch audio sample transform tests')
parser.add_argument ( '-s', '--size', type=int, default=DEFAULT_SAMPLE_SIZE, help='Size of audio to transform in frames')
parser.add_argument ('-p', '--period', type=int, default=DEFAULT_SAMPLE_PERIOD, help='Period size in frames')
parser.add_argument ('-t', '--trans', type=int, default=DEFAULT_TRANSFORM_SIZE, help='Transform size in bytes')
args = Test.parseArgs (parser)

Ipatch.init ()

# Determine host endian byte order
hostEndian = Ipatch.SampleEndian.BENDIAN if sys.byteorder == 'big' else Ipatch.SampleEndian.LENDIAN

# Create double floating point sample data audio
sampleSize = args.size
period = args.period
sampleDataArray = (ctypes.c_double * sampleSize)()

for pos in xrange (0, sampleSize):
  periodPos = math.fmod (pos, period) / period
  sampleDataArray[pos] = math.sin (periodPos * 2 * math.pi)

sampleData = bytearray (sampleDataArray)

# Create sample transform object and initialize size
trans = Ipatch.SampleTransform ()
trans.alloc_size (args.trans)

failcount = 0
max_maxdiff = 0.0

for isrc in xrange (0, len (testFormats)):
  srcform = testFormats[isrc]

  for idest in xrange (0, len (testFormats)):
    destform = testFormats[idest]

    # convert generated double floating point waveform to source format
    trans.set_formats (Ipatch.SampleWidth.DOUBLE | hostEndian, srcform,
                       Ipatch.SAMPLE_UNITY_CHANNEL_MAP)
    srcData = trans.convert (sampleData)

    # convert source format to destination format
    trans.set_formats (srcform, destform, Ipatch.SAMPLE_UNITY_CHANNEL_MAP)
    destData = trans.convert (srcData)

    # convert destination format to final double output
    trans.set_formats (destform, Ipatch.SampleWidth.DOUBLE | hostEndian,
                       Ipatch.SAMPLE_UNITY_CHANNEL_MAP)
    finalData = trans.convert (destData)
    finalDataArray = (ctypes.c_double * sampleSize).from_buffer_copy (finalData)

    maxdiff = 0.0
    maxindex = 0

    # compare final waveform against original
    for i in xrange (0, sampleSize):
      d = abs (sampleDataArray[i] - finalDataArray[i])

      if d > maxdiff:
        maxdiff = d
        maxindex = i

    if maxdiff > max_maxdiff:
      max_maxdiff = maxdiff

    if maxdiff <= MAX_DIFF_ALLOWED:
      Test.info ("Converted %s to %s: maxdiff=%0.6f, index=%d"
                 % (testFormatNames[isrc], testFormatNames[idest], maxdiff, maxindex))
    else:
      Test.error ("FAILED Converting %s to %s: maxdiff=%0.6f, index=%d"
                  % (testFormatNames[isrc], testFormatNames[idest], maxdiff, maxindex))
      failcount += 1

if failcount > 0:
  Test.error ("%d of %d format conversions FAILED (max diff=%0.6f)" % (failcount, len (testFormats) ** 2, max_maxdiff))
else: Test.msg ("All %d sample format conversions PASSED (max diff=%0.6f)" % (len (testFormats) ** 2, max_maxdiff))


Test.exit ()

